#include <gdk/gdkkeysyms.h>

#include "../gm-app.h"
#include "gm-world-view.h"
#include "gm-world-text-view.h"
#include "gm-world-input-view.h"
#include "gm-text-scroller.h"
#include "gm-editor-view.h"
#include "../gm-debug.h"
#include "../gm-support.h"
#include "../gm-color-table.h"
#include "../mcp/gm-mcp-package.h"
#include "../mcp/gm-mcp-session.h"
#include "../gm-editor.h"

#define GM_WORLD_VIEW_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE((object), \
		GM_TYPE_WORLD_VIEW, GmWorldViewPrivate))

struct _GmWorldViewPrivate {
	GmWorld *world;
	
	GtkStatusbar *statusbar;
	GtkHPaned *hpaned;
	GmWorldTextView *text_view_world;
	GmWorldInputView *text_view_input;
	GmTextScroller *text_scroller_world;
};

void on_gm_world_view_world_text_received(GmWorld *world, gchar *text, 
		GmWorldView *view);
void on_gm_world_view_world_error(GmWorld *world, gchar *text, gint code,
		GmWorldView *view);
void on_gm_world_input_view_world_text_activate(GmWorldInputView *iview, 
		gchar *text, GmWorldView *view);
void on_gm_world_view_world_state_changing(GmWorld *world, guint state, 
		GmWorldView *view);
void on_gm_world_view_world_active_changed(GmWorld *world, gboolean active,
		GmWorldView *view);

void on_gm_world_view_world_editor_added(GmWorld *world, GmEditor *editor,
		GmWorldView *view);
void on_gm_world_view_world_editor_removed(GmWorld *world, GmEditor *editor,
		GmWorldView *view);
		
void on_gm_world_view_world_mcp_package_created(GmMcpSession *session, 
		GmMcpPackage *package, GmWorldView *view);

void on_gm_world_view_editor_view_modified_changed(GmEditorView *editor_view, 
		gboolean modified, GmWorldView *view);
gboolean on_gm_world_view_world_text_view_scroll_event(GmWorldView *view, 
		GdkEventScroll *event, GmWorldTextView *text);
void on_gm_world_view_world_text_view_url_activate(GmWorldView *view,
		gchar const *url);

gboolean on_gm_world_view_world_input_view_key_pressed(GtkWidget *widget, 
		GdkEventKey *event, GmWorldView *view);
		
/* Signals */

/*enum {
  NUM_SIGNALS
};

static guint world_view_signals[NUM_SIGNALS] = {0};*/

G_DEFINE_TYPE(GmWorldView, gm_world_view, GTK_TYPE_NOTEBOOK)

static void
gm_world_view_finalize(GObject *object) {
	GmWorldView *view = GM_WORLD_VIEW(object);

	g_signal_handlers_disconnect_by_func(view->priv->world,
			G_CALLBACK(on_gm_world_view_world_text_received), view);
	g_signal_handlers_disconnect_by_func(view->priv->world,
			G_CALLBACK(on_gm_world_view_world_error), view);
	g_signal_handlers_disconnect_by_func(view->priv->world,
			G_CALLBACK(on_gm_world_view_world_state_changing), view);
	g_signal_handlers_disconnect_by_func(view->priv->world,
			G_CALLBACK(on_gm_world_view_world_active_changed), view);
	g_signal_handlers_disconnect_by_func(view->priv->world,
			G_CALLBACK(on_gm_world_view_world_editor_added), view);
	g_signal_handlers_disconnect_by_func(view->priv->world,
			G_CALLBACK(on_gm_world_view_world_editor_removed), view);

	g_signal_handlers_disconnect_by_func(
			gm_world_get_mcp_session(view->priv->world), 
			G_CALLBACK(on_gm_world_view_world_mcp_package_created), view);

	g_object_unref(view->priv->world);
	G_OBJECT_CLASS(gm_world_view_parent_class)->finalize(object);
}

static void
gm_world_view_class_init(GmWorldViewClass *klass) {
	GObjectClass *object_class = G_OBJECT_CLASS(klass);
	
	object_class->finalize = gm_world_view_finalize;
	
	g_type_class_add_private(object_class, sizeof(GmWorldViewPrivate));
}

GtkWidget *
gm_world_view_create_input_text_view(GmWorldView *view) {
	GtkWidget *scrolled_window = gtk_scrolled_window_new(NULL, NULL);
	GtkWidget *input_text_view = gm_world_input_view_new_with_color_table(
			gm_app_color_table(gm_app_instance()));
	
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
			GTK_POLICY_NEVER, GTK_POLICY_NEVER);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled_window),
			GTK_SHADOW_IN);
	gtk_container_add(GTK_CONTAINER(scrolled_window), input_text_view);
		
	view->priv->text_view_input = GM_WORLD_INPUT_VIEW(input_text_view);
	
	g_signal_connect(input_text_view, "key_press_event", 
			G_CALLBACK(on_gm_world_view_world_input_view_key_pressed), view);
	return scrolled_window;
}

GtkWidget *
gm_world_view_create_world_text_view(GmWorldView *view) {
	GtkWidget *world_text_view = gm_world_text_view_new_with_color_table(
			gm_app_color_table(gm_app_instance()));
	GtkWidget *scrolled_window = gtk_scrolled_window_new(NULL, NULL);
	
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled_window),
			GTK_SHADOW_IN);
	gtk_container_add(GTK_CONTAINER(scrolled_window), world_text_view);
	
	view->priv->text_view_world = GM_WORLD_TEXT_VIEW(world_text_view);
	
	// Create new text scroller, this object will take care of itself and will
	// destroy itself when the view dies, neat!
	view->priv->text_scroller_world = 
			gm_text_scroller_new(GTK_TEXT_VIEW(view->priv->text_view_world));
	
	g_signal_connect(world_text_view, "scroll_event",
			G_CALLBACK(on_gm_world_view_world_text_view_scroll_event), view);
	g_signal_connect(world_text_view, "url_activate",
			G_CALLBACK(on_gm_world_view_world_text_view_url_activate), view);

	return scrolled_window;
}

void
gm_world_view_update_status(GmWorldView *view, gchar const *status) {
	gtk_statusbar_pop(view->priv->statusbar, 0);
	
	if (status == NULL) {
		gtk_statusbar_push(view->priv->statusbar, 0, 
				_("Welcome to GnoeMoe, explorer of new worlds!"));
	} else {
		gtk_statusbar_push(view->priv->statusbar, 0, 
				status);	
	}
}

GmEditorView *
gm_world_view_editor_page_new(GmWorldView *view, GmEditor *editor) {
	GmEditorView *editor_view = gm_editor_view_new(view->priv->world, editor);
	GtkWidget *label;
	GmLabelInfo info;
	gchar const *icon;
	
	if (gm_editor_is_code(editor)) {
		icon = "editor_verb.xpm";
	} else {
		icon = "editor_text.xpm";
	}
	
	gtk_widget_show(GTK_WIDGET(editor_view));
	label = gm_create_tab_label("editor_verb.xpm", gm_editor_name(editor),
			TRUE, &info);
	gtk_notebook_append_page(GTK_NOTEBOOK(view), GTK_WIDGET(editor_view), 
			label);
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(view), TRUE);
	gtk_notebook_set_current_page(GTK_NOTEBOOK(view), 
			gtk_notebook_page_num(GTK_NOTEBOOK(view), 
			GTK_WIDGET(editor_view)));
	
	GM_DEBUG("Connecting modified changed!!!");
	
	g_signal_connect(editor_view, "modified-changed", 
			G_CALLBACK(on_gm_world_view_editor_view_modified_changed), 
			view);
	
	return editor_view;
}

GtkWidget *
gm_world_view_world_page_new(GmWorldView *view) {
	GtkWidget *vbox = gtk_vbox_new(FALSE, 0);
	GtkWidget *hpaned = gtk_hpaned_new();
	GtkWidget *status = gtk_statusbar_new();
	GtkWidget *vbox_world = gtk_vbox_new(FALSE, 3);

	gtk_box_pack_start(GTK_BOX(vbox_world), 
			gm_world_view_create_world_text_view(view), TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(vbox_world), 
			gm_world_view_create_input_text_view(view), FALSE, FALSE, 0);
	
	gtk_container_set_border_width(GTK_CONTAINER(vbox_world), 3);
	gtk_paned_pack1(GTK_PANED(hpaned), vbox_world, TRUE, TRUE);

	gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(status), FALSE);
	
	gtk_box_pack_start(GTK_BOX(vbox), hpaned, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), status, FALSE, FALSE, 0);

	view->priv->statusbar = GTK_STATUSBAR(status);
	view->priv->hpaned = GTK_HPANED(hpaned);

	gm_world_view_update_status(view, NULL);	
	
	return vbox;
}

static void
gm_world_view_init(GmWorldView *view) {
	GtkWidget *label;
	GmLabelInfo info;
	
	view->priv = GM_WORLD_VIEW_GET_PRIVATE(view);
	
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(view), TRUE);
	gtk_notebook_set_show_border(GTK_NOTEBOOK(view), FALSE);
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(view), GTK_POS_BOTTOM);
	gtk_notebook_set_scrollable(GTK_NOTEBOOK(view), TRUE);
	
	label = gm_create_tab_label("world.svg", _("World"), FALSE, &info);
	gtk_notebook_append_page(GTK_NOTEBOOK(view), 
			gm_world_view_world_page_new(view), label);
}

void
on_gm_world_view_destroy(GmWorldView *view, gpointer user_data) {
	gm_options_set_int(gm_world_options(view->priv->world), "pane_position",
			GTK_WIDGET(view->priv->hpaned)->allocation.width - 
			gtk_paned_get_position(GTK_PANED(view->priv->hpaned)));

	gm_options_save(gm_world_options(view->priv->world));
}

void
on_gm_world_view_show(GmWorldView *view, gpointer user_data) {
	gm_do_events();
	
    gtk_paned_set_position(GTK_PANED(view->priv->hpaned), 
    		GTK_WIDGET(view->priv->hpaned)->allocation.width
			- gm_options_get_int(gm_world_options(view->priv->world), 
			"pane_position"));
}

GtkWidget *
gm_world_view_new(GmWorld *world) {
	GmWorldView *view = GM_WORLD_VIEW(g_object_new(GM_TYPE_WORLD_VIEW, NULL));
	
	view->priv->world = g_object_ref(world);
	
	gm_world_input_view_set_history(view->priv->text_view_input, 
			gm_world_history(view->priv->world));

	g_signal_connect(world, "text_received", 
			G_CALLBACK(on_gm_world_view_world_text_received), view);
	g_signal_connect(world, "world_error", 
			G_CALLBACK(on_gm_world_view_world_error), view);
	g_signal_connect(world, "state_changing", 
			G_CALLBACK(on_gm_world_view_world_state_changing), view);
	g_signal_connect(world, "active_changed", 
			G_CALLBACK(on_gm_world_view_world_active_changed), view);
	g_signal_connect(world, "editor_added", 
			G_CALLBACK(on_gm_world_view_world_editor_added), view);
	g_signal_connect(world, "editor_removed", 
			G_CALLBACK(on_gm_world_view_world_editor_removed), view);

	g_signal_connect(gm_world_get_mcp_session(world), "package_created",
			G_CALLBACK(on_gm_world_view_world_mcp_package_created), view);

	g_signal_connect(view->priv->text_view_input, "text_activate", 
			G_CALLBACK(on_gm_world_input_view_world_text_activate), view);
	g_signal_connect(view, "destroy", 
			G_CALLBACK(on_gm_world_view_destroy), NULL);
	g_signal_connect(view, "show",
			G_CALLBACK(on_gm_world_view_show), NULL);

	return GTK_WIDGET(view);
}

gboolean
gm_world_view_find_first(GmWorldView *view, const gchar *str,
		GmWorldViewSearchFlags flags) {
	GtkTextView *tview = GTK_TEXT_VIEW(view->priv->text_view_world);
	GtkTextIter iter;
	GtkTextBuffer *buffer;
  
	if (tview) {  
		buffer = gtk_text_view_get_buffer(tview);
    
		if (buffer) {
			if (flags & GM_WORLD_VIEW_SEARCH_BACKWARDS) {
				gtk_text_buffer_get_end_iter(buffer, &iter);
			} else {
				gtk_text_buffer_get_start_iter(buffer, &iter);
			}
			
			gtk_text_buffer_place_cursor(buffer, &iter);
      
			return gm_world_view_find_next(view, str, flags);
		}
	}
  
	return FALSE;
}

gboolean
gm_world_view_find_next(GmWorldView *view, const gchar *str, 
		GmWorldViewSearchFlags flags) {
	GtkTextBuffer *buffer = NULL;
	GtkTextView *tview = NULL;
	GtkTextIter end, start, matchStart, matchEnd;
	gboolean found = FALSE;

	if (*str == '\0') {
		return FALSE;
	}

	tview = GTK_TEXT_VIEW(view->priv->text_view_world);

	if (tview) {
		buffer = gtk_text_view_get_buffer(tview);
	}

	if (buffer) {
		if (!gtk_text_buffer_get_selection_bounds(buffer, &start, &end)) {
			if (flags & GM_WORLD_VIEW_SEARCH_BACKWARDS) {
				gtk_text_buffer_get_end_iter(buffer, &end);
			} else {
				gtk_text_buffer_get_start_iter(buffer, &end);
			}
			
			start = end;
		}

		if (flags & GM_WORLD_VIEW_SEARCH_FORWARDS) {
			found = gtk_text_iter_forward_search(&end, str, 
					GTK_TEXT_SEARCH_VISIBLE_ONLY | GTK_TEXT_SEARCH_TEXT_ONLY, 
					&matchStart, &matchEnd, NULL);
		} else {
			found = gtk_text_iter_backward_search(&start, str,
					GTK_TEXT_SEARCH_VISIBLE_ONLY | GTK_TEXT_SEARCH_TEXT_ONLY, 
					&matchStart, &matchEnd, NULL);
		}

		if (found) {
			gtk_text_buffer_place_cursor(buffer, &matchStart);
			gtk_text_buffer_move_mark_by_name(buffer, "selection_bound", 
					&matchEnd);
			gtk_text_view_scroll_to_iter(tview, &matchStart, 
					0.0, FALSE, 0.0, 0.0);
		}
	}

	return found;
}

gboolean
gm_world_view_text_active(GmWorldView *view) {
	return gtk_notebook_get_current_page(GTK_NOTEBOOK(view)) == 0;
}

void
gm_world_view_scroll_end_prepare(GmWorldView *view) {
	// TODO
}

void
gm_world_view_scroll_end(GmWorldView *view) {
	// TODO
}

GmWorld *
gm_world_view_world(GmWorldView *view) {
	return view->priv->world;
}

GmWorldInputView *
gm_world_view_input(GmWorldView *view) {
	return view->priv->text_view_input;
}

GmWorldTextView *
gm_world_view_text_view(GmWorldView *view) {
	return view->priv->text_view_world;
}

GtkTextBuffer *
gm_world_view_buffer(GmWorldView *view) {
	return gtk_text_view_get_buffer(GTK_TEXT_VIEW(view->priv->text_view_world));
}

GtkHPaned *
gm_world_view_hpaned(GmWorldView *view) {
	return view->priv->hpaned;
}

void
gm_world_view_open_log(GmWorldView *view, const gchar *filename) {
	// TODO: open log
}

void
gm_world_view_set_userlist_width(GmWorldView *view, gint width) {
	gtk_paned_set_position(GTK_PANED(view->priv->hpaned), 
			GTK_WIDGET(view->priv->hpaned)->allocation.width - width);
}

void
gm_world_view_set_focus(GmWorldView *view) {
	gtk_widget_grab_focus(GTK_WIDGET(view->priv->text_view_input));
}

void
gm_world_view_change_font_size(GmWorldView *view, gint size_change) {
	GtkStyle *style = gtk_widget_get_style(GTK_WIDGET(view));
	PangoFontDescription *desc = style->font_desc;
	PangoFontDescription *copy = pango_font_description_copy(desc);
	gchar *new_font;
	
	pango_font_description_set_size(copy, 
			pango_font_description_get_size(copy) + 
			(size_change * PANGO_SCALE));
	new_font = pango_font_description_to_string(copy);
	gm_color_table_set_font_description(gm_app_color_table(gm_app_instance()),
			new_font);
	
	pango_font_description_free(copy);
	g_free(new_font);
}

/* Callbacks */
void
on_gm_world_input_view_world_text_activate(GmWorldInputView *iview, gchar *text,
		GmWorldView *view) {
	gm_world_process_input(view->priv->world, text);
	gm_text_scroller_scroll_end(view->priv->text_scroller_world);
}

void
on_gm_world_view_world_text_received(GmWorld *world, gchar *text, 
		GmWorldView *view) {
	gchar *inserted = 
			gm_world_text_view_insert(view->priv->text_view_world, text);

	g_free(inserted);
}

void
on_gm_world_view_editor_save(GmEditor *editor, GmWorldView *view) {
	gtk_notebook_set_current_page(GTK_NOTEBOOK(view), 0);
	gtk_widget_grab_focus(GTK_WIDGET(view->priv->text_view_input));
}

void
on_gm_world_view_editor_view_modified_changed(GmEditorView *editor_view, 
		gboolean modified, GmWorldView *view) {
	gint i, n = gtk_notebook_get_n_pages(GTK_NOTEBOOK(view));
	GtkWidget *page;
	GtkLabel *label;
	gchar *str;
	
	// Find this view then
	for (i = 1; i < n; ++i) {
		page = gtk_notebook_get_nth_page(GTK_NOTEBOOK(view), i);
		
		if (GM_IS_EDITOR_VIEW(page) && GM_EDITOR_VIEW(page) == editor_view) {
			label = GTK_LABEL(gm_container_item(GTK_CONTAINER(
					gtk_notebook_get_tab_label(GTK_NOTEBOOK(view), page)),
					GTK_TYPE_LABEL));
						
			if (!modified) {
				gtk_label_set_label(label, 
						gm_editor_name(gm_editor_view_editor(
						GM_EDITOR_VIEW(page))));
			} else {
				str = g_strconcat(gm_editor_name(gm_editor_view_editor(
						GM_EDITOR_VIEW(page))), "*", NULL);
				gtk_label_set_label(label, str);
				g_free(str);
			}
		}
	}
}

void
on_gm_world_view_world_editor_added(GmWorld *world, GmEditor *editor,
		GmWorldView *view) {
	gm_world_view_editor_page_new(view, editor);
	
	g_signal_connect(editor, "save", 
			G_CALLBACK(on_gm_world_view_editor_save), view);
}

void
on_gm_world_view_world_editor_removed(GmWorld *world, GmEditor *editor,
		GmWorldView *view) {
	gint i, n = gtk_notebook_get_n_pages(GTK_NOTEBOOK(view));
	GtkWidget *page;
	
	for (i = 1; i < n; ++i) {
		page = gtk_notebook_get_nth_page(GTK_NOTEBOOK(view), i);	
		
		if (GM_IS_EDITOR_VIEW(page) && gm_editor_view_editor(
				GM_EDITOR_VIEW(page)) == editor) {
			
			g_signal_handlers_disconnect_by_func(editor,
					on_gm_world_view_editor_save, view);
			gtk_widget_destroy(page);
			
			gtk_notebook_set_current_page(GTK_NOTEBOOK(view), 0);
			gtk_widget_grab_focus(GTK_WIDGET(view->priv->text_view_input));
		}
	}
}

void
on_gm_world_view_world_error(GmWorld *world, gchar *text, gint code,
		GmWorldView *view) {
	gchar *line;
	
	switch (code) {
		case GM_NET_ERROR_CONNECTING:
			line = g_strdup_printf(_("# Connect failed: %s"), text);
		break;
		case GM_NET_ERROR_DISCONNECTED:
			line = g_strdup_printf(_("# Connection lost... (%s)"), text);
		break;
		default:
			line = g_strdup_printf(_("# Error: %s"), text);
		break;
	}	

	gm_world_writeln(world, line);
	g_free(line);
}

void
on_gm_world_view_world_state_changing(GmWorld *world, guint state, 
		GmWorldView *view) {
	gchar *line = NULL;
	GmNetState pstate = gm_world_state(world);
	
	switch (state) {
		case GM_NET_STATE_TRY_ADDRESS:
			line = g_strdup_printf(_("# Trying %s port %s"), 
					gm_world_current_host(world), gm_world_current_port(world));
		break;
		case GM_NET_STATE_CONNECTING:
			line = g_strdup_printf(_("# Connecting to %s port %s"), 
					gm_world_current_host(world), gm_world_current_port(world));
		break;
		case GM_NET_STATE_CONNECTED:
			line = g_strdup(_("# Connected"));
		break;
		case GM_NET_STATE_DISCONNECTED:
			if (pstate == GM_NET_STATE_CONNECTED || 
					pstate == GM_NET_STATE_DISCONNECTING) {
				line = g_strdup(_("# Disconnected"));
			}
			gm_world_view_update_status(view, NULL);
		break;
		case GM_NET_STATE_DISCONNECTING:
			line = g_strdup(_("# Disconnecting"));
		break;
		default:
		break;
	}

	if (line) {	
		gm_world_writeln(world, line);
		g_free(line);
	}
}

void
on_gm_world_view_world_active_changed(GmWorld *world, gboolean active,
		GmWorldView *view) {
	if (active) {
		gtk_widget_grab_focus(GTK_WIDGET(view->priv->text_view_input));
	}
}

gboolean on_gm_world_view_world_text_view_scroll_event(GmWorldView *view, 
		GdkEventScroll *event, GmWorldTextView *text) {	
	if (event->state & GDK_CONTROL_MASK) {
		switch (event->direction) {
			case GDK_SCROLL_UP:
				// Decrease font size
				gm_world_view_change_font_size(view, -1);
			break;
			case GDK_SCROLL_DOWN:
				// Increase font size
				gm_world_view_change_font_size(view, 1);
			break;
			default:
			break;
		}
		return TRUE;
	} else {
		return FALSE;
	}
	
	return FALSE;
}

gboolean
on_gm_world_view_world_input_view_key_pressed(GtkWidget *widget, 
		GdkEventKey *event, GmWorldView *view) {
	switch (event->keyval) {
		case GDK_Home: case GDK_End:
			if ((event->state | GDK_CONTROL_MASK) == event->state) {
				if (event->keyval == GDK_End) {
					gm_text_scroller_scroll_end(
							view->priv->text_scroller_world);
				} else {
					gm_text_scroller_scroll_begin(
							view->priv->text_scroller_world);
				}
				
				return TRUE;
			}
		break;
		default:
		break;
	}
	
	return FALSE;
}

void
on_gm_world_view_world_mcp_package_created(GmMcpSession *session, 
		GmMcpPackage *package, GmWorldView *view) {
	gm_mcp_package_create_view(package, G_OBJECT(view));
}

void
on_gm_world_view_world_text_view_url_activate(GmWorldView *view,
		gchar const *url) {
	gm_open_url(url);
}
